/* -*-mb0-c-*-

   SB-PRAM simulator

   (C) 1994 by Michael Bosch (hirbli@cs.uni-sb.de)
   and Stefan Franziskus (stefran@cs.uni-sb.de)

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies.
*/

#include <stdio.h>
#include <stdarg.h>


enum io_class {IO_STDIO, IO_REG, IO_DISASS, IO_STDERR, IO_NOTICE, IO_DEBUG };

static int sim_print(enum io_class class, const char *format, ... )
{
	char buffer[4096];
	va_list arg;
	va_start(arg, format);

	IFOPT(OPT_XWIN)
	{
		switch(class)
		{
		case IO_STDIO:
			vsprintf(buffer, format, arg);
			return io_write(buffer);
			break;
		case IO_REG:
			vsprintf(buffer, format, arg);
			add_regs(buffer);
			return strlen(buffer);
			break;
		case IO_DISASS:
			vsprintf(buffer, format, arg);
			add_diss(buffer);
			break;
		case IO_STDERR:
			vsprintf(buffer, format, arg);
			sprintf(buffer, "\010%s\010", buffer);
			return notice_write(buffer);
			break;
		case IO_NOTICE:
			vsprintf(buffer, format, arg);
			return notice_write(buffer);
			break;
		case IO_DEBUG:
			vsprintf(buffer, format, arg);
			return notice_write(buffer);
			break;
		default:
			fprintf(stderr, "Oops wrong IO-class\n");          /* #### */
			return -1;
			break;
		}
	}
	else
		switch(class)
		{
		case IO_STDIO:
			return vprintf(format, arg);
			break;
		case IO_REG:
			return vprintf(format, arg);
			break;
		case IO_DISASS:
			return vprintf(format, arg);
			break;
		case IO_STDERR:
			return vfprintf(stderr, format, arg);
			break;
		case IO_NOTICE:
			return vfprintf(stderr, format, arg);
			break;
		case IO_DEBUG:
			return vprintf(format, arg);
			break;
		default:
			fprintf(stderr, "Oops wrong IO-class\n");          /* #### */
			return -1;
			break;
		}

	va_end(arg);
}
